class Config:
    MYSQL_HOST: str = "localhost"
    MYSQL_PORT: int = 3306
    MYSQL_DATABASE: str = ""
    MYSQL_USER: str = "mysql"
    MYSQL_PASSWORD: str = ""
    MYSQL_POOL_SIZE: int = None
    MYSQL_POOL_CONNECTION_TIMEOUT: int = 1000

    MANAGER_AUTO_MAP_MODEL_ATTRIBUTES = False
    """ If `True` => Model attributes will be mapped on class attributes automatically in results of `select_one` or `select_all` methods. """



import sys
import traceback
from datetime import datetime
from .cli_colors import CMD, FG


class BaseLogger:
    def __init__(self):
        self.format_string = "{datetime} [{prefix}] {event} | {kwargs}"

    def _log(self, prefix: str, event: str, **kwargs) -> str:
        """
        Overwrite this method with your own logger
        :param prefix: Log prefix
        :param event: Log event (critical. error, exception, info, warning, debug)
        :param kwargs: Custom keyword arguments
        """
        return self.format_string.format(prefix=prefix, event=event, datetime=datetime.now(), kwargs=kwargs)

    def critical(self, event: str, **kwargs) -> str:
        return self._log("critical", event, **kwargs)

    def error(self, event: str, **kwargs) -> str:
        return self._log("error", event, **kwargs)

    def exception(self, event: str, **kwargs) -> str:
        return self._log("exception", event, **kwargs)

    def info(self, event: str, **kwargs) -> str:
        return self._log("info", event, **kwargs)

    def warning(self, event: str, **kwargs) -> str:
        return self._log("warning", event, **kwargs)

    def debug(self, event: str, **kwargs) -> str:
        return self._log("debug", event, **kwargs)


class StdOutLogger(BaseLogger):
    def _log(self, prefix: str, event: str, **kwargs) -> None:
        print(prefix, event, file=sys.stdout)
        for key, value in kwargs.items():
            print(f"    {key}: {value}", file=sys.stdout)

    def critical(self, event: str, **kwargs) -> None:
        return self._log(f"{FG.red}{CMD.bold}[CRITICAL ]{CMD.reset} ", event, **kwargs)  # red/bold

    def error(self, event: str, **kwargs) -> None:
        return self._log(f"{FG.red}[ERROR    ]{CMD.reset} ", event, **kwargs)  # red

    def exception(self, event: str, **kwargs) -> None:
        traceback.print_exc(file=sys.stdout)
        return self._log(f"{FG.purple}{CMD.bold}[EXCEPTION]{CMD.reset} ", event, **kwargs)  # purple/bold

    def info(self, event: str, **kwargs) -> None:
        return self._log("[INFO     ] ", event, **kwargs)

    def warning(self, event: str, **kwargs) -> None:
        return self._log(f"{FG.orange}[WARNING  ]{CMD.reset} ", event, **kwargs)  # orange

    def debug(self, event: str, **kwargs) -> None:
        return self._log("[DEBUG    ] ", event, **kwargs)


class Logger:
    """
    Static class proxy for logger instance. Usage: Logger.log.[method]
    """
    log: BaseLogger = BaseLogger()
    """ Logger instance. Default is BaseLogger with no output into file or stdout. """

    @classmethod
    def set_external_logger(cls, logger_instance: BaseLogger) -> BaseLogger:
        """
        Set custom or predefined logger instance based on BaseLogger interface.
        :param logger_instance: Custom logger instance
        :return: Instance of setted logger
        """
        cls.log = logger_instance
        return cls.log



import os
import re
import sys

import sqlparse
from jinja2 import Template

from .datatypes import DATATYPES
from ..data_access.db import DBI
from ..tools.cli_colors import CMD, FG

class Analyser:
    def __init__(self, output_path: str, custom_templates_path: str = None):
        self.db = DBI()
        self.base_output_path = output_path
        self.models_output_path = os.path.join(self.base_output_path, "models")
        self.managers_output_path = os.path.join(self.base_output_path, "managers")

        self.template_path = custom_templates_path or os.path.join(os.path.dirname(__file__), os.path.dirname(__file__), "../templates/")
        self.model_template_path = os.path.join(self.template_path, "../templates/model.jinja")
        self.manager_template_path = os.path.join(self.template_path, "../templates/manager.jinja")
        self.manager_base_template_path = os.path.join(self.template_path, "../templates/manager_base.jinja")

        self.table_name: str = None
        self.table_type: str = None
        self.table_description: list = None
        self.attr_datatypes = None
        self.primary_keys: list = None
        self.model_imports: list = None
        self.model_convertors: dict = None
        self.enum_types: list = []
        self.view_statement: str = None
        self.view_statement_create: str = None
        self.where_base: str = None
        self.order_by_default: str = None

        self.model_j_template: Template = self._get_j_template_instance(self.model_template_path)
        self.manager_j_template: Template = self._get_j_template_instance(self.manager_template_path)
        self.manager_base_j_template: Template = self._get_j_template_instance(self.manager_base_template_path)

    def run(self):
        print(f"{FG.green}Starting Database Access Object Generator{CMD.reset}")
        table_list = self._list_tables()
        for table in table_list:
            try:
                self.table_name = table[0]
                self.table_type = table[1]
                self._describe_table()
                self._analyze_datatypes()
                if self.table_type == "VIEW":
                    self._parse_view_sql_statement()
                else:
                    self.view_statement = None
                    self.view_statement_create = None
                self._process_model_template()
                self._process_manager_template()
                self._process_manager_base_template()
            except Exception as ex:
                print(f"{FG.red}Error while processing table '{table}'{CMD.reset}: {ex.__str__()}", file=sys.stderr)
        print(f"{FG.green}DONE{CMD.reset}")

    def _list_tables(self):
        return self.db.fetch_all("SHOW FULL tables", dictionary_output=False)

    def _describe_table(self):
        self.table_description = self.db.fetch_all("SHOW FULL COLUMNS FROM `{}`".format(self.table_name))
        self.primary_keys = [item["Field"] for item in self.table_description if item["Key"] == "PRI"]

    def _parse_view_sql_statement(self):
        self.view_statement = None
        self.view_statement_create = None
        view_statement_info = self.db.fetch_one("SHOW CREATE TABLE `{}`".format(self.table_name))
        view_statement_create = view_statement_info["Create View"]
        found = re.findall(r"SECURITY DEFINER VIEW (`[^`]+`) AS (\(?.*\)?$)", view_statement_create)
        self.view_statement_create = sqlparse.format(
            view_statement_create, keyword_case="upper", indent_columns=True, wrap_after=2048
        )
        if found:
            _, view_statement = found[0]
            view_statement = view_statement[1:-1] if view_statement.startswith("(") else view_statement

            view_statement = sqlparse.format(
                view_statement, keyword_case="upper", indent_columns=True, wrap_after=32768
            )

            where_base = "1"
            where_regex = r"\nWHERE ([^\n]*)(\n|$)?"
            search_where = re.search(where_regex, view_statement)
            if search_where:
                where_base = search_where.groups()[0]
                view_statement = re.sub(where_regex, "\n", view_statement)

            order_by_default = ""
            order_by_regex = r"\nORDER BY ([^\n]*)(\n|$)?"
            search_order_by = re.search(order_by_regex, view_statement)
            if search_order_by:
                order_by_default = search_order_by.groups()[0]
                view_statement = re.sub(order_by_regex, "\n", view_statement)

            self.where_base = where_base
            self.order_by_default = order_by_default

            view_statement = sqlparse.format(view_statement, keyword_case="upper", indent_columns=True)

            search_group_by = re.search(r"\nGROUP BY", view_statement)
            if search_group_by:
                found_on_index = search_group_by.start()
                view_statement_parts = (view_statement[:found_on_index], view_statement[found_on_index:])
                view_statement_template = f"{view_statement_parts[0]} \n{{WHERE}} {view_statement_parts[1]} \n{{ORDER_BY}} \n{{LIMIT}} \n{{OFFSET}}"  # noqa
            else:
                view_statement_template = f"{view_statement}\n{{WHERE}} \n{{ORDER_BY}} \n{{LIMIT}} \n{{OFFSET}}"

            self.view_statement = view_statement_template

    def _process_model_template(self):
        model_name = self._get_model_name(self.table_name)
        model_filename = "{}_model.py".format(self.table_name.lower())

        output = self.model_j_template.render(
            modelName=model_name,
            tableName=self.table_name,
            tableType=self.table_type,
            tableDescription=self.table_description,
            modelImports=self.model_imports,
            modelConvertors=self.model_convertors,
            primaryKeys=self.primary_keys,
            enumTypes=self.enum_types,
            dataTypes=self.attr_datatypes,
            viewStatement=self.view_statement,
            viewStatementCreate=self.view_statement_create,
            whereBase=self.where_base,
            orderByDefault=self.order_by_default,
        )

        if self.base_output_path:
            self._create_module(self.base_output_path)
            self._create_module(self.models_output_path)
            file_path = os.path.join(self.models_output_path, model_filename)
            print(f"{CMD.bold}Writing model{CMD.reset} `{model_name}` into `{file_path}`")
            with open(file_path, "w") as f:
                f.write(output)
        else:
            print(f"**** MODEL: {model_name} --> {model_filename} ****\n{output}")

    def _process_manager_base_template(self):
        manager_name = self._get_model_name(self.table_name)
        manager_filename = "{}_manager_base.py".format(self.table_name.lower())

        output = self.manager_base_j_template.render(
            modelName=manager_name,
            tableName=self.table_name,
            tableType=self.table_type,
            primaryKeys=self.primary_keys,
            dataTypes=self.attr_datatypes,
        )

        if self.managers_output_path:
            self._create_module(os.path.join(self.managers_output_path, "base"))
            file_path = os.path.join(self.managers_output_path, "base", manager_filename)
            print(f"{CMD.bold}Writing base manager{CMD.reset} `{manager_name}` into `{file_path}`")
            with open(file_path, "w") as f:
                f.write(output)
        else:
            print(f"**** BASE MANAGER: {manager_name} --> {manager_filename} ****\n{output}")

    def _process_manager_template(self):
        manager_name = self._get_model_name(self.table_name)
        manager_filename = "{}_manager.py".format(self.table_name.lower())

        output = self.manager_j_template.render(modelName=manager_name, tableName=self.table_name)
        if self.base_output_path:
            self._create_module(self.base_output_path)
            self._create_module(self.managers_output_path)
            file_path = os.path.join(self.managers_output_path, manager_filename)
            if os.path.exists(file_path):
                print(f"Skipping manager `{manager_name}` exists `{file_path}`")
            else:
                print(f"{CMD.bold}Writing manager{CMD.reset} `{manager_name}` into `{file_path}`")
                with open(file_path, "w") as f:
                    f.write(output)
        else:
            print(f"**** MANAGER: {manager_name} --> {manager_filename} ****\n{output}")

    def _analyze_datatypes(self):
        self.model_imports = []
        self.model_convertors = {}
        self.enum_types = []
        self.attr_datatypes = {}
        for item in self.table_description:
            attr_name = item["Field"]
            model_datatype = None
            db_datatype, db_datatype_size = self._get_db_datatype(item["Type"])
            model_datatype_info = DATATYPES.get(
                db_datatype, DATATYPES["__default__"]
            )
            if db_datatype not in DATATYPES:
                print(f"{FG.orange}WARNING:{CMD.reset} Unknown datatype '{item['Type']}' found in table '{self.table_name}' attribute '{item['Field']}'. "
                      f"It will converted into string by default. Best way is to add it into DATATYPES dict.")
            if model_datatype_info:
                model_datatype = model_datatype_info[0]
                if model_datatype_info[1]:
                    self.model_imports.append(model_datatype_info[1])
                if model_datatype_info[2]:
                    self.model_convertors[attr_name] = model_datatype_info[2]
                # process ENUMs
                if db_datatype == "enum":
                    enums_str = db_datatype_size.replace("'", '"').replace(',"', ', "')
                    self.enum_types.append({"Field": attr_name, "Options": enums_str})
            item["ModelType"] = model_datatype
            item["ModelDefaultValue"] = self._wrap_word(item["Default"], model_datatype_info[3], model_datatype_info[4])
            self.attr_datatypes[attr_name] = model_datatype
        self.model_imports = list(set(self.model_imports))
        self.model_imports.sort()

    @staticmethod
    def _get_j_template_instance(template_path: str) -> Template:
        with open(template_path, "r") as f:
            template = f.read()
        j_template = Template(template)
        return j_template

    @staticmethod
    def _get_model_name(table_name: str) -> str:
        spl = table_name.lower().split("_")
        spl = map(str.capitalize, spl)
        return "".join(spl)

    @staticmethod
    def _get_db_datatype(db_datatype) -> (str, str):
        found = re.findall(r"(^[a-zA-Z0-9_]+)(\([^)]*\))?", db_datatype)
        _db_datatype = found[0][0] if found else None
        _db_datatype_size = found[0][1][1:-1] if found and len(found[0][1]) > 2 else None
        return _db_datatype, _db_datatype_size

    @staticmethod
    def _wrap_word(word: str, left_wrapper: str, right_wrapper: str) -> str:
        return "{}{}{}".format(left_wrapper or "", word, right_wrapper or "") if word else None

    @staticmethod
    def _create_module(module_path: str):
        try:
            os.makedirs(module_path)
            init_file = os.path.join(module_path, "__init__.py")
            open(init_file, "a").close()
        except Exception as ex:
            #print(f"_create_module exception: {ex}")
            pass



import typing


class ModelBase:
    """
    Base model class
    """

    class Meta:
        TABLE_NAME: str = None
        TABLE_TYPE: str = None
        SQL_STATEMENT: str = None
        SQL_STATEMENT_WHERE_BASE: str = None
        SQL_STATEMENT_ORDER_BY_DEFAULT: str = ""
        PRIMARY_KEYS: typing.List = []
        ATTRIBUTE_LIST: typing.List = []
        ATTRIBUTE_TYPES: typing.Dict = {}
        MODEL_DATA_CONVERTOR: typing.Dict = {}

    DATATYPES_CONVERTOR = {"<class 'decimal.Decimal'>": float}

    def __init__(self, init_data: typing.Dict = {}):
        self.model_data: typing.Dict = self._convert_datatypes(init_data)

    def __str__(self):
        return str(self.model_data)

    def __setattr__(self, key, value):
        if key in self.Meta.ATTRIBUTE_LIST and hasattr(self, "model_data"):
            self.model_data[key] = value
        return super().__setattr__(key, value)

    def to_dict(self) -> typing.Dict:
        """
        Returns internal model data as dict
        """
        return self.model_data

    def map_model_attributes(self, data: typing.Dict = None) -> "ModelBase":
        """
        Set or update model attributes by internal model data or external data from method param if attribute exists.
        :param data: External data to be mapped
        """
        if data is None:
            data = self.model_data
        for key, value in data.items():
            if key in self.Meta.ATTRIBUTE_LIST:
                self.__setattr__(key, value)
        return self

    def clone(self):
        model_clone = self.__class__()
        for attribute_name in self.Meta.ATTRIBUTE_LIST:
            model_clone.__setattr__(attribute_name, self.__getattribute__(attribute_name))
            model_clone.model_data = self.model_data.copy()
        return model_clone

    @classmethod
    def _convert_datatypes(cls, item: dict) -> dict:
        for key, value in item.items():
            datatype = str(type(value))
            # known problematic data type needed conversion for easy jSON transform
            if datatype in cls.DATATYPES_CONVERTOR:
                item[key] = cls.DATATYPES_CONVERTOR[datatype](value)
            # data convertion for DB type
            if key in cls.Meta.MODEL_DATA_CONVERTOR:
                item[key] = cls.Meta.MODEL_DATA_CONVERTOR[key](value)
        return item



import typing
from functools import wraps
from time import sleep

from mysql.connector import Error
from mysql.connector import MySQLConnection
from mysql.connector.pooling import MySQLConnectionPool
from mysql.connector.pooling import errors
from ..tools.log import Logger
from ..config import Config


class DBI:
    is_initialized = False
    connection_config: dict = None
    connection_pool: MySQLConnectionPool = None

    def __init__(self):
        self._is_in_transaction = False
        self._is_in_pass_dbi = False
        self._is_in_self_dbi = False
        self._connection = None
        DBI._init()

    @classmethod
    def _init(cls):
        if cls.is_initialized:
            return
        cls.connection_config = {
            "host": Config.MYSQL_HOST,
            "port": Config.MYSQL_PORT,
            "database": Config.MYSQL_DATABASE,
            "user": Config.MYSQL_USER,
            "password": Config.MYSQL_PASSWORD,
        }
        cls.connection_pool = (
            MySQLConnectionPool(
                pool_name="app_pool",
                pool_size=Config.MYSQL_POOL_SIZE,
                pool_reset_session=False,
                **cls.connection_config
            )
            if Config.MYSQL_POOL_SIZE
            else None
        )

    def _get_connection(self):
        if not self._connection:
            if DBI.connection_pool:
                try_get_pool_conn_counter = Config.MYSQL_POOL_CONNECTION_TIMEOUT
                while not self._connection:
                    try:
                        if not try_get_pool_conn_counter:
                            raise Exception(
                                "Pool connection timeout error. No connection avaliable in pool during {}ms".format(
                                    Config.MYSQL_POOL_CONNECTION_TIMEOUT
                                )
                            )
                        self._connection = DBI.connection_pool.get_connection()
                    except errors.PoolError:
                        Logger.log.debug(
                            "DBI._get_connection.wait_for_pool_connection: {}".format(try_get_pool_conn_counter)
                        )
                        try_get_pool_conn_counter -= 1
                        sleep(0.001)
            else:
                self._connection = MySQLConnection(**DBI.connection_config)
            Logger.log.debug(
                "DBI._get_connection.new",
                connection_id=self._connection.connection_id,
                pooled=DBI.connection_pool is not None,
            )
        return self._connection

    def execute(self, sql: str, sql_args: typing.Tuple = ()) -> int:
        """
        For executing CRUD SQL commands.
        INSERT returns last inserted ID
        UPDATE, DELETE returns number of affected rows
        :param sql: SQL command
        :param sql_args: Tuple of positioned SQL arguments. It will safely replace "%s" sequences.
        :return: Number of affexted rows or last inserted ID or False if command failed.
        """
        Logger.log.debug("DBI.execute", sql=sql, sql_args=sql_args)

        ret = False
        cursor = None
        is_insert_command = sql.upper().startswith("INSERT")
        try:
            if self._get_connection().is_connected():
                cursor = self._get_connection().cursor()
                Logger.log.debug("DBI.execute.execute")
                cursor.execute(sql, sql_args)
                self._commit()

                ret = cursor.lastrowid if is_insert_command else cursor.rowcount
        except Error as ex:
            Logger.log.exception("DBI.execute.exception", message=ex)
            raise ex
        finally:
            if self._get_connection().is_connected():
                if cursor:
                    cursor.close()

                self._close_connection()
        return ret

    def execute_many(self, sql: str, sql_args: typing.List[typing.Tuple]) -> int:
        """
        For executing many CRUD SQL commands (Bulk inserts).
        INSERT returns last inserted ID
        UPDATE, DELETE returns number of affected rows
        :param sql: SQL command
        :param sql_args: List of tuples with positional SQL arguments. It will safely replace "%s" sequences.
        :return: Number of affexted rows or last inserted ID or False if command failed.
        """
        Logger.log.debug("DBI.execute_many", sql=sql, sql_args=sql_args)

        ret = False
        cursor = None
        try:
            if self._get_connection().is_connected():
                cursor = self._get_connection().cursor()
                Logger.log.debug("DBI.execute_many.executemany")
                cursor.executemany(sql, sql_args)
                self._commit()

                ret = cursor.rowcount
        except Error as ex:
            Logger.log.exception("DBI.execute_many", message=ex)
            raise ex
        finally:
            if self._get_connection().is_connected():
                if cursor:
                    cursor.close()

                self._close_connection()
        return ret

    def fetch_one(self, sql, sql_args: tuple = (), dictionary_output=True) -> typing.Dict:
        record = None
        cursor = None
        try:
            if self._get_connection().is_connected():
                cursor = self._get_connection().cursor(dictionary=dictionary_output)
                Logger.log.debug("DBI.fetch_one", sql=sql)
                cursor.execute(sql, sql_args)
                record = cursor.fetchone()
        except Error as ex:
            Logger.log.exception("DBI.fetch_one", message=ex)
            raise ex
        finally:
            if self._get_connection().is_connected():
                if cursor:
                    cursor.close()

                self._close_connection()
        return record

    def fetch_all(self, sql, sql_args: tuple = (), dictionary_output=True) -> typing.List[typing.Dict]:
        records = None
        cursor = None
        try:
            if self._get_connection().is_connected():
                cursor = self._get_connection().cursor(dictionary=dictionary_output)
                Logger.log.debug("DBI.fetch_all", sql=sql)
                cursor.execute(sql, sql_args)
                records = cursor.fetchall()

        except Error as ex:
            Logger.log.exception("DBI.fetch_all", message=ex)
            raise ex
        finally:
            if self._get_connection().is_connected():
                if cursor:
                    cursor.close()

                self._close_connection()
        return records

    @classmethod
    def use_self_dbi(cls, dbi_attr_name: str = "dbi"):
        """
        Allow decorated function to work with one instance of DB connection
        :param dbi_attr_name: Name of attribute where DBI instance is stored in self. "dbi" is default.
        """

        def decorator(fnc):
            @wraps(fnc)
            def wrapper(*args, **kwargs):
                class_instance = args[0]
                dbi = class_instance.__getattribute__(dbi_attr_name)
                dbi._is_in_self_dbi = True
                Logger.log.debug("DBI.use_self_dbi.start")
                try:
                    ret = fnc(*args, **kwargs)
                except Exception as ex:
                    raise ex
                finally:
                    dbi._is_in_self_dbi = False
                    Logger.log.debug("DBI.use_self_dbi.done")
                    dbi._close_connection()
                return ret

            return wrapper

        return decorator

    @classmethod
    def pass_dbi(cls, pass_dbi_as: str = "dbi"):
        """
        Pass DBI instasnce into wrapped method
        :param pass_dbi_as: Name of argument for passing DBI instance. "dbi" is default.
        """

        def decorator(fnc):
            @wraps(fnc)
            def wrapper(*args, **kwargs):
                dbi = cls()
                dbi._is_in_pass_dbi = True
                Logger.log.debug("DBI.pass_dbi.start")
                try:
                    kwargs[pass_dbi_as] = dbi
                    ret = fnc(*args, **kwargs)
                except Exception as ex:
                    raise ex
                finally:
                    dbi._is_in_pass_dbi = False
                    Logger.log.debug("DBI.pass_dbi.done")
                    dbi._close_connection()
                return ret

            return wrapper

        return decorator

    @classmethod
    def transaction(cls, pass_dbi_as: str = "dbi"):
        """
        Transaction wrapper

        How to use it:
            @DBI.transaction("dbi")\n
            def run(self, dbi):\n
                dbi.execute("INSERT INTO `table1` (`column1`, `column2`) VALUES ('value1", 44)")\n
                dbi.execute("INSERT INTO `table2` (`column1`, `column2`) VALUES ('value2", 55)")\n
                dbi.execute("INSERT INTO `table3` (`column1`, `column2`) VALUES ('value3", 66)")\n

        How to use it with DataManagers:
            @DBI.transaction("dbi")\n
            def run(self, dbi):\n
                model1 = DataManager.create_model_instance()\n
                model1.number = 14\n
                manager = RegionRegionsManager(dbi=dbi)\n
                manager.insert_one(model1)\n
                model2 = DataManager.create_model_instance()\n
                model2.number = 15\n
                manager.insert_one(mod)\n

        :param pass_dbi_as: Name of argument for passing DBI instance. "dbi" is default.
        """

        def decorator(fnc):
            @wraps(fnc)
            def wrapper(*args, **kwargs):
                dbi = cls()
                dbi._is_in_transaction = True
                Logger.log.debug("DBI.transaction.start")
                dbi._get_connection().start_transaction()
                try:
                    kwargs[pass_dbi_as] = dbi
                    ret = fnc(*args, **kwargs)
                except Exception as ex:
                    Logger.log.exception("DBI.transaction.rollback", message=ex)
                    dbi._get_connection().rollback()
                    raise
                else:
                    dbi._is_in_transaction = False
                    dbi._commit()
                finally:
                    dbi._is_in_transaction = False
                    Logger.log.debug("DBI.transaction.done")
                    dbi._close_connection()
                return ret

            return wrapper

        return decorator

    def _commit(self):
        if not self._is_in_transaction:
            Logger.log.debug("DBI._commit")
            self._connection.commit()
            return True

        return False

    def _close_connection(self):
        if self._is_in_transaction or self._is_in_pass_dbi or self._is_in_self_dbi:
            return True
        try:
            Logger.log.debug("DBI._close_connection", connection_id=self._connection.connection_id)
            self._connection.close()
            self._connection = None
            return True
        except Error as ex:
            Logger.log.exception("DBI._close_connection", message=ex)
        self._connection = None

        return False



import typing

from ..tools.log import Logger

from .db import DBI
from .model_base import ModelBase
from ..config import Config


class ManagerException(BaseException):
    pass


class ViewManagerBase:
    MODEL_CLASS = ModelBase

    def __init__(self, dbi: DBI = None):
        """
        Init function of base model manager class
        :param dbi: Instance of database connector. If empty it will be created automatically. Instance of DBI is usualy used with combination of transaction wrapper @DBI.transaction("dbi")
        """
        self.dbi = DBI() if dbi is None else dbi
        self.bulk_insert_buffer_size = 50
        self.bulk_insert_sql_statement = ""
        self.bulk_insert_values_buffer = []

    @classmethod
    def create_model_instance(cls, init_data: dict = None) -> ModelBase:
        if init_data is None:
            init_data = {}

        return cls.MODEL_CLASS(init_data)

    def select_one(
        self,
        *args,
        condition: str = "1",
        condition_params: typing.Tuple = (),
        projection: typing.Tuple = (),
        order_by: typing.Tuple = (),
    ) -> ModelBase:
        """
        Select one row from DB table or View
        :param projection: sql projection - default *
        :param args: Primary keys or condition and condition_params if there are no primary keys
        :param condition: SQL Condition (Will be used if there are no positional args from primary keys)
        :param condition_params: Positional params for SQL condition
            (Will be used if there are no positional args from primary keys)
        :param order_by: Params for SQL order by statement
        """
        base_condition = self.MODEL_CLASS.Meta.SQL_STATEMENT_WHERE_BASE

        if args:
            condition = self._prepare_primary_sql_condition()
            condition_params = args

        projection_statement = ", ".join(projection) if projection else "*"
        order_by_sql_format = ", ".join(order_by)
        limit = 1

        if base_condition == "1":
            where_statement = f"WHERE ({condition})" if condition else ""
        else:
            where_statement = f"WHERE {base_condition} AND ({condition})" if condition else f"WHERE {base_condition}"

        order_by_statement = f"ORDER BY {order_by_sql_format}" if order_by else ""
        limit_statement = f"LIMIT {limit}" if limit else ""

        sql = self.MODEL_CLASS.Meta.SQL_STATEMENT.format(
            PROJECTION=projection_statement,
            WHERE=where_statement,
            ORDER_BY=order_by_statement,
            LIMIT=limit_statement,
            OFFSET="",
        )

        Logger.log.info("ViewManagerBase.select_one.sql", manager=self.__class__.__name__)

        result = self.dbi.fetch_one(sql, condition_params)

        Logger.log.info("ViewManagerBase.select_one.result", result=result, manager=self.__class__.__name__)

        if Config.MANAGER_AUTO_MAP_MODEL_ATTRIBUTES:
            return self.MODEL_CLASS(result).map_model_attributes() if result else None

        return self.MODEL_CLASS(result) if result else None

    def select_all(
        self,
        condition: str = "1",
        condition_params: typing.Tuple = (),
        projection: typing.Tuple = (),
        order_by: typing.Tuple = (),
        limit: int = 0,
        offset: int = 0,
    ) -> typing.List[ModelBase]:
        """
        Select all rows matching the condition
        :param offset: SQL offset
        :param projection: sql projection - default *
        :param condition: SQL condition
        :param condition_params: Positional params for SQL condition
        :param order_by: Params for SQL order by statement
        :param limit: Params for SQL limit statement
        """
        base_condition = self.MODEL_CLASS.Meta.SQL_STATEMENT_WHERE_BASE

        projection_statement = ", ".join(projection) if projection else "*"

        if base_condition == "1":
            where_statement = f"WHERE ({condition})" if condition else ""
        else:
            where_statement = f"WHERE {base_condition} AND ({condition})" if condition else f"WHERE {base_condition}"

        order_by_sql_format = ", ".join(order_by)
        if len(order_by) > 0:
            order_by_statement = f"ORDER BY {order_by_sql_format}"
        else:
            if self.MODEL_CLASS.Meta.SQL_STATEMENT_ORDER_BY_DEFAULT:
                order_by_statement = f"ORDER BY {self.MODEL_CLASS.Meta.SQL_STATEMENT_ORDER_BY_DEFAULT}"
            else:
                order_by_statement = ""

        limit_statement = f"LIMIT {limit}" if limit else ""
        offset_statement = f"OFFSET {offset}" if offset else ""

        sql = self.MODEL_CLASS.Meta.SQL_STATEMENT.format(
            PROJECTION=projection_statement,
            WHERE=where_statement,
            ORDER_BY=order_by_statement,
            LIMIT=limit_statement,
            OFFSET=offset_statement,
        )

        Logger.log.info("ViewManagerBase.select_all.sql", manager=self.__class__.__name__)

        results = self.dbi.fetch_all(sql, condition_params)

        Logger.log.info("ViewManagerBase.select_all.result", result=results, manager=self.__class__.__name__)

        if Config.MANAGER_AUTO_MAP_MODEL_ATTRIBUTES:
            Logger.log.debug("ViewManagerBase.select_all.result.list.automapped")
            return [self.MODEL_CLASS(result).map_model_attributes() for result in results]

        Logger.log.debug("ViewManagerBase.select_all.result.list")
        return [self.MODEL_CLASS(result) for result in results]

    @staticmethod
    def models_into_dicts(result: typing.List[ModelBase]) -> typing.List[typing.Dict]:
        """
        Convert result of select_all into list of dicts
        :param result: List of models
        """
        return [item.to_dict() for item in result]

    @classmethod
    def _prepare_primary_sql_condition(cls):
        args = ["{} = %s".format(primary_key) for primary_key in cls.MODEL_CLASS.Meta.PRIMARY_KEYS]
        return " AND ".join(args)

    @classmethod
    def _prepare_primary_sql_condition_params(cls, model_instance: ModelBase):
        return [model_instance.__getattribute__(attribute_name) for attribute_name in cls.MODEL_CLASS.Meta.PRIMARY_KEYS]


class TableManagerBase(ViewManagerBase):
    def update_one(self, model_instance: ModelBase) -> int:
        """
        Update one database record based on model attributes
        :param model_instance: Model instance
        :return: Number of affected rows
        """
        if not self.MODEL_CLASS.Meta.PRIMARY_KEYS:
            raise ManagerException("Can't update record based on model instance. There are no primary keys specified.")

        set_prepare = []
        set_prepare_params = []
        for attribute_name in self.MODEL_CLASS.Meta.ATTRIBUTE_LIST:
            if model_instance.model_data.get(attribute_name) != model_instance.__getattribute__(attribute_name):
                set_prepare.append("`{}` = %s".format(attribute_name))
                set_prepare_params.append(model_instance.__getattribute__(attribute_name))

        if len(set_prepare) == 0:  # no update needed
            return 0

        condition_prepare = self._prepare_primary_sql_condition()
        condition_prepare_params = self._prepare_primary_sql_condition_params(model_instance)

        sql = "UPDATE `{}` SET {} WHERE {} LIMIT 1".format(
            self.MODEL_CLASS.Meta.TABLE_NAME, ", ".join(set_prepare), condition_prepare
        )

        Logger.log.info("TableManagerBase.update_one.sql", manager=self.__class__.__name__)

        result = self.dbi.execute(sql, set_prepare_params + condition_prepare_params)

        Logger.log.info("TableManagerBase.update_one.result", result=result, manager=self.__class__.__name__)

        return result

    def insert_one(
        self,
        model_instance: ModelBase,
        use_on_duplicate_update_statement: bool = False,
        use_insert_ignore_statement: bool = False,
    ) -> int:
        """
        Insert one record into database based on model attributes
        :param model_instance: Model instance
        :param use_on_duplicate_update_statement: Use ON DUPLICATE KEY UPDATE statement
        :param use_insert_ignore_statement: Use INSERT IGNORE statement
        :return: Last inserted id if it is possible
        """
        insert_prepare = []
        insert_prepare_values = []
        insert_prepare_params = []
        update_prepare = []
        for attribute_name in self.MODEL_CLASS.Meta.ATTRIBUTE_LIST:
            insert_prepare.append("`{}`".format(attribute_name))
            insert_prepare_values.append("%s")
            insert_prepare_params.append(model_instance.__getattribute__(attribute_name))
            if use_on_duplicate_update_statement:
                update_prepare.append("`{0}` = VALUES(`{0}`)".format(attribute_name))

        if use_on_duplicate_update_statement:
            sql = "INSERT INTO `{}` ({}) VALUES ({}) ON DUPLICATE KEY UPDATE {}".format(
                self.MODEL_CLASS.Meta.TABLE_NAME,
                ", ".join(insert_prepare),
                ", ".join(insert_prepare_values),
                ", ".join(update_prepare),
            )
        elif use_insert_ignore_statement:
            sql = "INSERT IGNORE INTO `{}` ({}) VALUES ({})".format(
                self.MODEL_CLASS.Meta.TABLE_NAME, ", ".join(insert_prepare), ", ".join(insert_prepare_values)
            )
        else:
            sql = "INSERT INTO `{}` ({}) VALUES ({})".format(
                self.MODEL_CLASS.Meta.TABLE_NAME, ", ".join(insert_prepare), ", ".join(insert_prepare_values)
            )

        Logger.log.info("TableManagerBase.insert_one.sql", manager=self.__class__.__name__)

        result = self.dbi.execute(sql, insert_prepare_params)

        # set primary key value
        if (
            result
            and len(self.MODEL_CLASS.Meta.PRIMARY_KEYS) == 1
            and self.MODEL_CLASS.Meta.ATTRIBUTE_TYPES[self.MODEL_CLASS.Meta.PRIMARY_KEYS[0]] == int
        ):
            model_instance.__setattr__(self.MODEL_CLASS.Meta.PRIMARY_KEYS[0], result)

        Logger.log.info("TableManagerBase.insert_one.result", result=result, manager=self.__class__.__name__)

        return result

    def insert_one_bulk(
        self,
        model_instance: ModelBase,
        use_on_duplicate_update_statement: bool = False,
        use_insert_ignore_statement: bool = False,
        auto_flush: bool = True,
    ) -> int:
        """
        Insert more records in one bulk.
        :param model_instance: Model instance
        :param use_on_duplicate_update_statement: Use ON DUPLICATE KEY UPDATE statement
        :param use_insert_ignore_statement: Use INSERT IGNORE statement
        :param auto_flush: Auto flush bulks from buffer after N records (defined in self.bulk_insert_buffer_size)
        :return: Number of items in buffer
        """
        insert_prepare = []
        insert_prepare_values = []
        insert_prepare_params = []
        update_prepare = []
        for attribute_name in self.MODEL_CLASS.Meta.ATTRIBUTE_LIST:
            insert_prepare.append("`{}`".format(attribute_name))
            insert_prepare_values.append("%s")
            insert_prepare_params.append(model_instance.__getattribute__(attribute_name))
            if use_on_duplicate_update_statement:
                update_prepare.append("`{0}` = VALUES(`{0}`)".format(attribute_name))

        if not self.bulk_insert_sql_statement:
            if use_on_duplicate_update_statement:
                self.bulk_insert_sql_statement = "INSERT INTO `{}` ({}) VALUES ({}) ON DUPLICATE KEY UPDATE {}".format(
                    self.MODEL_CLASS.Meta.TABLE_NAME,
                    ", ".join(insert_prepare),
                    ", ".join(insert_prepare_values),
                    ", ".join(update_prepare),
                )
            elif use_insert_ignore_statement:
                self.bulk_insert_sql_statement = "INSERT IGNORE INTO `{}` ({}) VALUES ({})".format(
                    self.MODEL_CLASS.Meta.TABLE_NAME, ", ".join(insert_prepare), ", ".join(insert_prepare_values)
                )
            else:
                self.bulk_insert_sql_statement = "INSERT INTO `{}` ({}) VALUES ({})".format(
                    self.MODEL_CLASS.Meta.TABLE_NAME, ", ".join(insert_prepare), ", ".join(insert_prepare_values)
                )

        self.bulk_insert_values_buffer.append(insert_prepare_params)
        buffer_len = len(self.bulk_insert_values_buffer)
        if auto_flush and buffer_len >= self.bulk_insert_buffer_size:
            self.insert_bulk_flush()

        return buffer_len

    def insert_bulk_flush(self) -> int:
        """
        Flush prepared inserts from buffer
        :return: Number of inserted rows
        """

        result = None
        if self.bulk_insert_values_buffer:
            result = self.dbi.execute_many(self.bulk_insert_sql_statement, self.bulk_insert_values_buffer)

        Logger.log.info(
            "TableManagerBase.insert_one_bulk_flush.result",
            result=result,
            inserted_count=len(self.bulk_insert_values_buffer),
            manager=self.__class__.__name__,
        )

        self.bulk_insert_sql_statement = ""
        self.bulk_insert_values_buffer = []
        return result

    def delete_one(self, model_instance: ModelBase) -> int:
        """
        Delete one row matching primary key condition.
        :param model_instance: Instance of model
        :return: Number of affected rows
        """
        condition_prepare = self._prepare_primary_sql_condition()
        condition_prepare_params = self._prepare_primary_sql_condition_params(model_instance)

        sql_statement = "DELETE FROM  `{}` WHERE {} LIMIT 1"
        sql = sql_statement.format(self.MODEL_CLASS.Meta.TABLE_NAME, condition_prepare)

        Logger.log.info("TableManagerBase.delete_one.sql", manager=self.__class__.__name__)

        result = self.dbi.execute(sql, condition_prepare_params)

        Logger.log.info(f"TableManagerBase.delete_one.result", result=result, manager=self.__class__.__name__)

        return result

    def delete_all(
        self, condition: str, condition_params: typing.Tuple = (), order_by: typing.Tuple = (), limit: int = 0
    ) -> int:
        """
        Delete all table rows matching condition.
        :param condition: SQL condition statement
        :param condition_params: SQL condition position params
        :param order_by: SQL order statement
        :param limit: SQL limit statement
        :return: Number of affected rows
        """
        where_statement = f"WHERE {condition}"
        order_by_sql_format = ", ".join(order_by)
        order_by_statement = f"ORDER BY {order_by_sql_format}" if order_by else ""
        limit_statement = f"LIMIT {limit}" if limit else ""

        sql_statement = "DELETE FROM `{TABLE}` {WHERE} {ORDER_BY} {LIMIT}"
        sql = sql_statement.format(
            TABLE=self.MODEL_CLASS.Meta.TABLE_NAME,
            WHERE=where_statement,
            ORDER_BY=order_by_statement,
            LIMIT=limit_statement,
        )

        Logger.log.info("TableManagerBase.delete_all.sql", manager=self.__class__.__name__)

        result = self.dbi.execute(sql, condition_params)

        Logger.log.info("TableManagerBase.delete_all.result", result=result, manager=self.__class__.__name__)

        return result



DATATYPES = {
    "__default__": ("str", None, "str", '"', '"'),
    "bigint": ("int", None, None, "", ""),
    "bit": ("int", None, None, "", ""),
    "blob": ("bytes", None, None, '"', '"'),
    "mediumblob": ("bytes", None, None, '"', '"'),
    "enum": ("str", None, None, '"', '"'),
    "float": ("float", None, None, "", ""),
    "decimal": ("float", None, "float", "", ""),
    "date": ("datetime.datetime", "datetime", None, 'datetime.datetime.strptime("', '", "%Y-%m-%d")'),
    "datetime": ("datetime.datetime", "datetime", None, 'datetime.datetime.strptime("', '", "%Y-%m-%d %H:%M:%S")'),
    "double": ("float", None, None, "", ""),
    "int": ("int", None, None, "", ""),
    "smallint": ("int", None, None, "", ""),
    "text": ("str", None, None, '"', '"'),
    "mediumtext": ("str", None, None, '"', '"'),
    "timestamp": ("datetime.datetime", "datetime", None, "None  # ", ""),
    "varbinary": ("str", None, None, '"', '"'),
    "varchar": ("str", None, None, '"', '"'),
    "tinyint": ("int", None, None, "", ""),
}
"""
Datatypes structure:
"DBDataType": (
    ModelDataType,
    NeededImport,
    NeededPostConvertFunction,
    LeftWrapperForDefaultValue,
    RightWrapperForDefaultValue)
Because jSON serializer can't easily convert all database types for example 'decimal', it should be converted into float by postprocess on Model side.
"""


import os
import stat
import sys

DEFAULT_MYSQL_PORT = 3306


def _exit(msg):
    print(msg)
    sys.exit(1)


def _save_configuration(_options, _output_path: str):
    file_content = f"#!/usr/bin/env bash\nszndaogen -a {_options.db_host} -r {_options.db_port} -d {_options.db_name} -u {_options.db_user} -p {_options.db_pass} {_output_path}\n"  # noqa
    file_name = f"szndaogen-{_options.db_host}-{_options.db_name}.sh"
    with open(file_name, "w+") as file:
        file.write(f"{file_content}\n")

        st = os.stat(file_name)
        os.chmod(file_name, st.st_mode | stat.S_IEXEC)

        print(f"\nShortcut script '{file_name}' created in current working directory.")
        print(f"File contents: {file_content}")


def wizard(options) -> tuple:
    continue_ = input("\nRequired parameters are not satisfied. Would you like to run setup wizard? [Y/n] ")
    if not(continue_ == "" or continue_.lower() == "y"):
        sys.exit(0)

    host = input("MySQL host address: ")
    if not host:
        _exit("No MySQL host provided. Exiting.")

    port = input("MySQL port (default 3306): ")
    if not port:
        port = DEFAULT_MYSQL_PORT

    db = input("MySQL database name: ")
    if not db:
        _exit("No MySQL DB name provided. Exiting.")

    user = input("MySQL username: ")
    if not user:
        _exit("No MySQL username provided. Exiting.")

    pw = input("MySQL password: ")

    output_path = input("Output path where all models and managers will be generated (default \"./data_access\"): ")
    if not output_path:
        output_path = "./data_access"

    options.db_host = host
    options.db_port = int(port)
    options.db_name = db
    options.db_user = user
    options.db_pass = pw

    arguments = [output_path]

    save_to_file = input("Before you proceed, would you like to save this configuration as a bash script in CWD for future use? [Y/n]")  # noqa
    if save_to_file == "" or save_to_file.lower() == "y":
        _save_configuration(options, output_path)

    return options, arguments



import pytest

from .auto_group import auto_group_dict
from .auto_group import auto_group_list
from .auto_group import auto_group_list_by_pkeys

_input1 = [
    {"a": 1, "b": 2, "c__a": 33, "c__b": 44},
    {"a": 1, "b": 2, "c__a": 55, "c__b": 66},
    {"a": 2, "b": 2, "c__a": 7, "c__b": 88},
    {"a": 2, "b": 2, "c__a": 77, "c__b": 99},
]
_expected1_1 = {
    "1": {"a": 1, "b": 2, "c": [{"a": 33, "b": 44}, {"a": 55, "b": 66}]},
    "2": {"a": 2, "b": 2, "c": [{"a": 7, "b": 88}, {"a": 77, "b": 99}]},
}
_expected1_2 = {
    "1-2": {"a": 1, "b": 2, "c": [{"a": 33, "b": 44}, {"a": 55, "b": 66}]},
    "2-2": {"a": 2, "b": 2, "c": [{"a": 7, "b": 88}, {"a": 77, "b": 99}]},
}

auto_group_list_by_pkeys_input = [(("a",), _input1, _expected1_1), (("a", "b"), _input1, _expected1_2)]


@pytest.mark.parametrize("primary_key_names, input, expected", auto_group_list_by_pkeys_input)
def test_auto_group_list_by_pkeys(primary_key_names, input, expected):
    assert auto_group_list_by_pkeys(primary_key_names, input) == expected


_input2 = [
    {"a": 1, "b": 2, "c__a": 3, "c__b": 4},
    {"a": 1, "b": 2, "c__a": 5, "c__b": 6},
    {"a": 1, "b": 2, "c__a": 7, "c__b": 8},
]
_expected2 = {"a": 1, "b": 2, "c": [{"a": 3, "b": 4}, {"a": 5, "b": 6}, {"a": 7, "b": 8}]}
auto_group_list_input = [(_input2, _expected2)]


@pytest.mark.parametrize("input, expected", auto_group_list_input)
def test_auto_group_list(input, expected):
    assert auto_group_list(input) == expected


_input3 = {"a": 1, "b": 2, "c___a": 3, "c___b___bb": 4, "c___b___cc": 5}
_expected3 = {"a": 1, "b": 2, "c": {"a": 3, "b": {"bb": 4, "cc": 5}}}

auto_group_dict_input = [(_input3, _expected3)]


@pytest.mark.parametrize("input, expected", auto_group_dict_input)
def test_auto_group_dict(input, expected):
    assert auto_group_dict(input) == expected



#!/usr/bin/python3
from setuptools import find_packages, setup
from szndaogen.tools.setuptools import get_file_content, get_file_content_as_list

packages = find_packages()
print(f"Found packages: {packages}")
VERSION = get_file_content("szndaogen/VERSION")
INSTALL_REQUIRES = get_file_content_as_list("requirements.txt")
DOCUMENTATION_MD = get_file_content("README.md")

setup(
    name="szndaogen",
    version=VERSION,
    author="Ales Adamek, Filip Cima, Richard Paprok",
    author_email="hpo.sport@firma.seznam.cz",
    description="SZN Database Access Object Generator for MySQL. Generates Models and DataManagers from existing MySQL database structure.",
    long_description=DOCUMENTATION_MD,
    long_description_content_type="text/markdown",
    url="https://github.com/seznam/szndaogen",
    packages=packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=INSTALL_REQUIRES,
    include_package_data=True,  # MANIFEST.in
    zip_safe=False,  # aby se spravne vycitala statika pridana pomoci MANIFEST.in
    entry_points={"console_scripts": ["szndaogen=szndaogen.cli:main"]},
)



import os

from optparse import OptionParser

from szndaogen.cli_wizard import wizard
from szndaogen.config import Config
from szndaogen.generator.analyser import Analyser
from szndaogen.tools.log import Logger, BaseLogger
from szndaogen.tools.setuptools import get_file_content


def get_cmd_options() -> tuple:
    version = get_file_content(os.path.join(os.path.dirname(__file__), "VERSION"))
    parser = OptionParser()
    parser.usage = (
        f"SZN DAOGen v{version} - Database Access Object Generator by Seznam.cz\n"
        "%prog [options] output_path\n"
        "    example: %prog -a localhost -d my_database -u root -p pass /path/to/data_access\n"
        "    example: %prog -a localhost -d my_database -u root -p pass ./data_access\n"
    )

    parser.add_option("-a", "--host-address", dest="db_host", type="string", help="MySQL database host. (required)")
    parser.add_option("-r", "--port", dest="db_port", type="int", help="MySQL database port.", default=3306)
    parser.add_option("-d", "--database", dest="db_name", type="string", help="MySQL database name. (required)")
    parser.add_option(
        "-u", "--user", dest="db_user", type="string", help="User name for MySQL DB authentication. (required)"
    )
    parser.add_option("-p", "--password", dest="db_pass", type="string", help="Password for MySQL DB authentication.")
    parser.add_option("-t", "--templates-path", dest="templates_path", type="string", help="Path to custom templates of Models (model.jinja), "
                                                                                           "DataManagers (manager.jinja) and DataManagerBases (manager_base.jinja).")

    options, arguments = parser.parse_args()

    if not options.db_host or not options.db_name or not options.db_user:
        parser.print_help()
        options, arguments = wizard(options)

    if len(arguments) < 1:
        print("Warning: Missing output path parameters. Models and Managers will be generated into std-out.")

    return options, arguments


def main():
    _options, _arguments = get_cmd_options()
    output_path = None if len(_arguments) < 1 else _arguments[0]

    Config.MYSQL_HOST = _options.db_host
    Config.MYSQL_PORT = _options.db_port
    Config.MYSQL_DATABASE = _options.db_name
    Config.MYSQL_USER = _options.db_user
    Config.MYSQL_PASSWORD = _options.db_pass
    Logger.set_external_logger(logger_instance=BaseLogger())

    app = Analyser(output_path, custom_templates_path=_options.templates_path)
    app.run()


if __name__ == "__main__":
    main()



import typing

from .model_base import ModelBase


class TModel(ModelBase):
    class Meta:
        TABLE_NAME: str = "table"
        TABLE_TYPE: str = "BASE TABLE"
        # fmt: off
        SQL_STATEMENT: str = "SELECT {PROJECTION} FROM `table` {WHERE} {ORDER_BY} {LIMIT} {OFFSET}"
        # fmt: on

        SQL_STATEMENT_WHERE_BASE: str = "1"
        SQL_STATEMENT_ORDER_BY_DEFAULT: str = ""

        PRIMARY_KEYS: typing.List = ["id", ]
        ATTRIBUTE_LIST: typing.List = ["id", "name", ]
        ATTRIBUTE_TYPES: typing.Dict = {
            "id": int,
            "name": str,
        }
        MODEL_DATA_CONVERTOR: typing.Dict = {
        }

    def __init__(self, init_data: typing.Dict = {}):
        super().__init__(init_data)
        self.id: int = None
        """Type: int(11), Can be NULL: NO, Key: PRI"""
        self.name: str = None
        """Type: varchar(50), Can be NULL: NO"""


def test_model_dataflow():
    model = TModel()
    model.id = 1
    model.name = "total_name"
    assert model.to_dict() == {"id": 1, "name": "total_name"}


def test_model_map_model_attributes():
    model = TModel()
    model.map_model_attributes({"id": 1, "name": "total_name"})
    assert model.id == 1
    assert model.name == "total_name"


def test_model_clone():
    model = TModel()
    model.id = 1
    model.name = "total_name"

    model_clone = model.clone()
    assert model_clone.id == model.id
    assert model_clone.name == model.name
    assert model_clone.to_dict() == model.to_dict()
